@@ -94,14 +94,14 @@ static struct {
 	char hostname[HOST_NAME_MAX+1];
 	bool reopening_logs;
 	bool schedule_reopen_logs;
+	int forced_log_priority;
+	bool disable_syslog;
 
 	struct debug_settings settings;
 	debug_callback_fn callback;
 	void *callback_private;
 	char header_str[300];
-	char header_str_no_nl[300];
 	size_t hs_len;
-	char msg_no_nl[FORMAT_BUFR_SIZE];
 } state = {
 	.settings = {
 		.timestamp_logs = true
@@ -230,6 +230,10 @@ static int debug_level_to_priority(int level)
 	};
 	int priority;
 
+	if (state.forced_log_priority != -1) {
+		level = state.forced_log_priority;
+	}
+
 	if (level < 0 || (size_t)level >= ARRAY_SIZE(priority_map))
 		priority = LOG_DEBUG;
 	else
@@ -239,48 +243,6 @@ static int debug_level_to_priority(int level)
 }
 #endif
 
-/* -------------------------------------------------------------------------- **
- * Produce a version of the given buffer without any trailing newlines.
- */
-#if defined(HAVE_LIBSYSTEMD_JOURNAL) || defined(HAVE_LIBSYSTEMD) || \
-	defined(HAVE_LTTNG_TRACEF) || defined(HAVE_GPFS)
-static void copy_no_nl(char *out,
-		       size_t out_size,
-		       const char *in,
-		       size_t in_len)
-{
-	size_t len;
-	/*
-	 * Some backends already add an extra newline, so also provide
-	 * a buffer without the newline character.
-	 */
-	len = MIN(in_len, out_size - 1);
-	if ((len > 0) && (in[len - 1] == '\n')) {
-		len--;
-	}
-
-	memcpy(out, in, len);
-	out[len] = '\0';
-}
-
-static void ensure_copy_no_nl(char *out,
-			      size_t out_size,
-			      const char *in,
-			      size_t in_len)
-{
-	/*
-	 * Assume out is a static buffer that is reused as a cache.
-	 * If it isn't empty then this has already been done with the
-	 * same input.
-	 */
-	if (out[0] != '\0') {
-		return;
-	}
-
-	copy_no_nl(out, out_size, in, in_len);
-}
-#endif
-
 /* -------------------------------------------------------------------------- **
  * Debug backends. When logging to DEBUG_FILE, send the log entries to
  * all active backends.
@@ -341,6 +303,10 @@ static void debug_syslog_log(int msg_level, const char *msg, size_t msg_len)
 {
 	int priority;
 
+	if (state.disable_syslog) {
+		return;
+	}
+
 	priority = debug_level_to_priority(msg_level);
 
 	/*
@@ -361,24 +327,33 @@ static void debug_syslog_log(int msg_level, const char *msg, size_t msg_len)
 static void debug_systemd_log(int msg_level, const char *msg, size_t msg_len)
 {
 	if (state.hs_len > 0) {
-		ensure_copy_no_nl(state.header_str_no_nl,
-				  sizeof(state.header_str_no_nl),
-				  state.header_str,
-				  state.hs_len);
-		sd_journal_send("MESSAGE=%s",
-				state.header_str_no_nl,
+		size_t len = state.hs_len;
+
+		if (state.header_str[len - 1] == '\n') {
+			len -= 1;
+		}
+
+		sd_journal_send("MESSAGE=%.*s",
+				(int)len,
+				state.header_str,
 				"PRIORITY=%d",
 				debug_level_to_priority(msg_level),
 				"LEVEL=%d",
 				msg_level,
 				NULL);
 	}
-	ensure_copy_no_nl(state.msg_no_nl,
-			  sizeof(state.msg_no_nl),
-			  msg, msg_len);
-	sd_journal_send("MESSAGE=%s", state.msg_no_nl,
-			"PRIORITY=%d", debug_level_to_priority(msg_level),
-			"LEVEL=%d", msg_level,
+
+	if ((msg_len > 0) && (msg[msg_len - 1] == '\n')) {
+		msg_len -= 1;
+	}
+
+	sd_journal_send("MESSAGE=%.*s",
+			(int)msg_len,
+			msg,
+			"PRIORITY=%d",
+			debug_level_to_priority(msg_level),
+			"LEVEL=%d",
+			msg_level,
 			NULL);
 }
 #endif
@@ -388,16 +363,19 @@ static void debug_systemd_log(int msg_level, const char *msg, size_t msg_len)
 static void debug_lttng_log(int msg_level, const char *msg, size_t msg_len)
 {
 	if (state.hs_len > 0) {
-		ensure_copy_no_nl(state.header_str_no_nl,
-				  sizeof(state.header_str_no_nl),
-				  state.header_str,
-				  state.hs_len);
-		tracef(state.header_str_no_nl);
+		size_t len = state.hs_len;
+
+		if (state.header_str[len - 1] == '\n') {
+			len -= 1;
+		}
+
+		tracef("%.*s", (int)len, state.header_str);
 	}
-	ensure_copy_no_nl(state.msg_no_nl,
-			  sizeof(state.msg_no_nl),
-			  msg, msg_len);
-	tracef(state.msg_no_nl);
+
+	if ((msg_len > 0) && (msg[msg_len - 1] == '\n')) {
+		msg_len -= 1;
+	}
+	tracef("%.*s", (int)msg_len, msg);
 }
 #endif /* WITH_LTTNG_TRACEF */
 
@@ -406,7 +384,9 @@ static void debug_lttng_log(int msg_level, const char *msg, size_t msg_len)
 static void debug_gpfs_reload(bool enabled, bool previously_enabled,
 			      const char *prog_name, char *option)
 {
-	gpfswrap_init();
+	if (enabled) {
+		gpfswrap_init();
+	}
 
 	if (enabled && !previously_enabled) {
 		gpfswrap_init_trace();
@@ -426,19 +406,39 @@ static void debug_gpfs_reload(bool enabled, bool previously_enabled,
 	}
 }
 
+static void copy_no_nl(char *out,
+		       size_t out_size,
+		       const char *in,
+		       size_t in_len)
+{
+	size_t len;
+	/*
+	 * Some backends already add an extra newline, so also provide
+	 * a buffer without the newline character.
+	 */
+	len = MIN(in_len, out_size - 1);
+	if ((len > 0) && (in[len - 1] == '\n')) {
+		len--;
+	}
+
+	memcpy(out, in, len);
+	out[len] = '\0';
+}
+
 static void debug_gpfs_log(int msg_level, const char *msg, size_t msg_len)
 {
+	char no_nl[FORMAT_BUFR_SIZE];
+
 	if (state.hs_len > 0) {
-		ensure_copy_no_nl(state.header_str_no_nl,
-				  sizeof(state.header_str_no_nl),
-				  state.header_str,
-				  state.hs_len);
-		gpfswrap_add_trace(msg_level, state.header_str_no_nl);
+		copy_no_nl(no_nl,
+			   sizeof(no_nl),
+			   state.header_str,
+			   state.hs_len);
+		gpfswrap_add_trace(msg_level, no_nl);
 	}
-	ensure_copy_no_nl(state.msg_no_nl,
-			  sizeof(state.msg_no_nl),
-			  msg, msg_len);
-	gpfswrap_add_trace(msg_level, state.msg_no_nl);
+
+	copy_no_nl(no_nl, sizeof(no_nl), msg, msg_len);
+	gpfswrap_add_trace(msg_level, no_nl);
 }
 #endif /* HAVE_GPFS */
 
@@ -693,13 +693,6 @@ static void debug_backends_log(const char *msg, size_t msg_len, int msg_level)
 {
 	size_t i;
 
-	/*
-	 * Some backends already add an extra newline, so initialize a
-	 * buffer without the newline character.  It will be filled by
-	 * the first backend that needs it.
-	 */
-	state.msg_no_nl[0] = '\0';
-
 	for (i = 0; i < ARRAY_SIZE(debug_backends); i++) {
 		if (msg_level <= debug_backends[i].log_level) {
 			debug_backends[i].log(msg_level, msg, msg_len);
@@ -1131,6 +1124,21 @@ void debug_set_hostname(const char *name)
 	strlcpy(state.hostname, name, sizeof(state.hostname));
 }
 
+void debug_set_forced_log_priority(int forced_log_priority)
+{
+	state.forced_log_priority = forced_log_priority;
+}
+
+void debug_disable_syslog(void)
+{
+	state.disable_syslog = true;
+}
+
+void debug_enable_syslog(void)
+{
+	state.disable_syslog = false;
+}
+
 /**
  * Ensure debug logs are initialised.
  *
@@ -1289,7 +1297,7 @@ bool reopen_logs_internal(void)
 	struct debug_backend *b = NULL;
 	mode_t oldumask;
 	size_t i;
-	bool ok;
+	bool ok = true;
 
 	if (state.reopening_logs) {
 		return true;
@@ -1568,13 +1576,18 @@ static void Debug1(const char *msg, size_t msg_len)
 	case DEBUG_STDERR:
 	case DEBUG_DEFAULT_STDOUT:
 	case DEBUG_DEFAULT_STDERR:
-		if (dbgc_config[DBGC_ALL].fd > 0) {
-			ssize_t ret;
-			do {
-				ret = write(dbgc_config[DBGC_ALL].fd,
-					    msg,
-					    msg_len);
-			} while (ret == -1 && errno == EINTR);
+		if (state.settings.debug_syslog_format ==
+		    DEBUG_SYSLOG_FORMAT_ALWAYS) {
+			debug_file_log(current_msg_level, msg, msg_len);
+		} else {
+			if (dbgc_config[DBGC_ALL].fd > 0) {
+				ssize_t ret;
+				do {
+					ret = write(dbgc_config[DBGC_ALL].fd,
+						    msg,
+						    msg_len);
+				} while (ret == -1 && errno == EINTR);
+			}
 		}
 		break;
 	case DEBUG_FILE:
@@ -1598,6 +1611,17 @@ static void bufr_print( void )
 	format_pos = 0;
 }
 
+/*
+ * If set (by tevent_thread_call_depth_set()) to value > 0, debug code will use
+ * it for the trace indentation.
+ */
+static size_t debug_call_depth = 0;
+
+size_t *debug_call_depth_addr(void)
+{
+	return &debug_call_depth;
+}
+
 /***************************************************************************
  Format the debug message text.
 
@@ -1624,8 +1648,21 @@ static void format_debug_text( const char *msg )
 	for( i = 0; msg[i]; i++ ) {
 		/* Indent two spaces at each new line. */
 		if(timestamp && 0 == format_pos) {
+			/* Limit the maximum indentation to 20 levels */
+			size_t depth = MIN(20, debug_call_depth);
 			format_bufr[0] = format_bufr[1] = ' ';
 			format_pos = 2;
+			/*
+			 * Indent by four spaces for each depth level,
+			 * but only if the current debug level is >= 8.
+			 */
+			if (depth > 0 && debuglevel_get() >= 8 &&
+			    format_pos + 4 * depth < FORMAT_BUFR_SIZE) {
+				memset(&format_bufr[format_pos],
+				       ' ',
+				       4 * depth);
+				format_pos += 4 * depth;
+			}
 		}
 
 		/* If there's room, copy the character to the format buffer. */
@@ -1725,23 +1762,31 @@ bool dbghdrclass(int level, int cls, const char *location, const char *func)
 
 	dbgsetclass(level, cls);
 
-	/* Don't print a header if we're logging to stdout. */
-	if ( state.logtype != DEBUG_FILE ) {
-		return( true );
+	/*
+	 * Don't print a header if we're logging to stdout,
+	 * unless 'debug syslog format = always'
+	 */
+	if (state.logtype != DEBUG_FILE &&
+	    state.settings.debug_syslog_format != DEBUG_SYSLOG_FORMAT_ALWAYS)
+	{
+		return true;
 	}
 
-	/* Print the header if timestamps are turned on.  If parameters are
-	 * not yet loaded, then default to timestamps on.
+	/*
+	 * Print the header if timestamps (or debug syslog format) is
+	 * turned on.  If parameters are not yet loaded, then default
+	 * to timestamps on.
 	 */
 	if (!(state.settings.timestamp_logs ||
 	      state.settings.debug_prefix_timestamp ||
-	      state.settings.debug_syslog_format)) {
+	      state.settings.debug_syslog_format != DEBUG_SYSLOG_FORMAT_NO))
+	{
 		return true;
 	}
 
 	GetTimeOfDay(&tv);
 
-	if (state.settings.debug_syslog_format) {
+	if (state.settings.debug_syslog_format != DEBUG_SYSLOG_FORMAT_NO) {
 		if (state.settings.debug_hires_timestamp) {
 			timeval_str_buf(&tv, true, true, &tvbuf);
 		} else {
@@ -1841,6 +1886,16 @@ bool dbghdrclass(int level, int cls, const char *location, const char *func)
 		}
 	}
 
+	if (debug_call_depth > 0) {
+		state.hs_len += snprintf(state.header_str + state.hs_len,
+					 sizeof(state.header_str) - state.hs_len,
+					 ", depth=%zu",
+					 debug_call_depth);
+		if (state.hs_len >= sizeof(state.header_str) - 1) {
+			goto full;
+		}
+	}
+
 	state.header_str[state.hs_len] = ']';
 	state.hs_len++;
 	if (state.hs_len < sizeof(state.header_str) - 1) {
@@ -1872,8 +1927,6 @@ full:
 		state.hs_len = sizeof(state.header_str) - 1;
 	}
 
-	state.header_str_no_nl[0] = '\0';
-
 	errno = old_errno;
 	return( true );
 }
@@ -1929,12 +1982,12 @@ static uint64_t debug_traceid = 0;
 
 uint64_t debug_traceid_set(uint64_t id)
 {
-    uint64_t old_id = debug_traceid;
-    debug_traceid = id;
-    return old_id;
+	uint64_t old_id = debug_traceid;
+	debug_traceid = id;
+	return old_id;
 }
 
 uint64_t debug_traceid_get(void)
 {
-    return debug_traceid;
+	return debug_traceid;
 }
